"""Update package data."""

import time

from pyhelpers.ops import confirmed

from .collector import LineData, OtherAssets
from .parser import get_site_map
from .utils import is_homepage_connectable, print_connection_warning


def _update_prepacked_data(verbose=False, interval=5, **kwargs):
    # noinspection PyUnresolvedReferences
    """
    Updates pre-packed data.

    :param verbose: Whether to print relevant information to the console; defaults to ``True``.
    :type verbose: bool | int
    :param interval: A time gap (in seconds) between updating different classes; defaults to ``5``.
    :type interval: int or float

    **Examples**::

        >>> from pyrcs._updater import _update_prepacked_data
        >>> _update_prepacked_data(verbose=True)
    """

    if not is_homepage_connectable():
        print_connection_warning(verbose=verbose)
        print("Unable to update the data.")

    else:
        if confirmed("To update the backup/data resources\n?", **kwargs):

            # Site map
            print("\nSite map:")
            _ = get_site_map(update=True, confirmation_required=False, verbose=verbose)

            time.sleep(interval)

            # Line data
            ld = LineData(update=True)
            ld.update(confirmation_required=False, verbose=verbose, interval=interval)

            time.sleep(interval)

            # Other assets
            oa = OtherAssets(update=True)
            oa.update(confirmation_required=False, verbose=verbose, interval=interval)

            if verbose:
                print("\nUpdate finished.")



"""
Collect data of `signal box prefix codes <http://www.railwaycodes.org.uk/signal/signal_boxes0.shtm>`_.
"""

import collections
import string
import urllib.parse

import bs4
import pandas as pd

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import get_collect_verbosity_for_fetch, homepage_url, is_homepage_connectable, \
    print_instance_connection_error, print_void_collection_message, validate_initial


class SignalBoxes(_Base):
    """
    A class for collecting data of
    `signal box prefix codes <http://www.railwaycodes.org.uk/signal/signal_boxes0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Signal box prefix codes'
    #: The key for accessing the data.
    KEY: str = 'Signal boxes'

    #: The key for accessing the data of *non-national rail*.
    KEY_TO_NON_NATIONAL_RAIL: str = 'Non-National Rail'
    #: The key for accessing the data of *Ireland*.
    KEY_TO_IRELAND: str = 'Ireland'
    #: The key for accessing the data of
    #: *WR (Western region) MAS (multiple aspect signalling) dates*.
    KEY_TO_WRMASD: str = 'WR MAS dates'
    #: The key for accessing the data of *bell codes*.
    KEY_TO_BELL_CODES: str = 'Bell codes'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/signal/signal_boxes0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb.NAME
            'Signal box prefix codes'
            >>> sb.URL
            'http://www.railwaycodes.org.uk/signal/signal_boxes0.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="other-assets",
            update=update, verbose=verbose)

    def _collect_prefix_codes(self, initial, source, verbose=False):
        initial_ = validate_initial(initial=initial)

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')
        thead, tbody = soup.find('thead'), soup.find('tbody')

        ths = [th.get_text(strip=True) for th in thead.find_all('th')]
        trs = tbody.find_all('tr')
        signal_boxes_data_table = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        signal_box_prefix_codes = {
            initial_: signal_boxes_data_table,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup)
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=signal_box_prefix_codes, data_name=initial_, sub_dir="a-z", verbose=verbose)

        return signal_box_prefix_codes

    def collect_prefix_codes(self, initial, confirmation_required=True, verbose=False,
                             raise_error=False):
        """
        Collects `signal box prefix codes`_ for a given initial letter from the source web page.

        .. _`signal box prefix codes`: http://www.railwaycodes.org.uk/signal/signal_boxes0.shtm

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of signal box prefix code.
        :type initial: str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing data of signal box prefix codes whose initial letters are
            the specified ``initial`` and the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_a_codes = sb.collect_prefix_codes(initial='a')
            >>> type(sb_a_codes)
            dict
            >>> list(sb_a_codes.keys())
            ['A', 'Last updated date']
            >>> sb_a_codes_dat = sb_a_codes['A']
            >>> type(sb_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> sb_a_codes_dat.head()
              Code               Signal Box  ...            Closed        Control to
            0   AF  Abbey Foregate Junction  ...
            1   AJ           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            2    R           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            3   AW               Abbey Wood  ...      13 July 1975      Dartford (D)
            4   AE         Abbey Works East  ...   1 November 1987  Port Talbot (PT)
            [5 rows x 8 columns]
        """

        initial_ = validate_initial(initial=initial)

        signal_box_prefix_codes = self._collect_data_from_source(
            data_name=self.NAME.lower(), method=self._collect_prefix_codes, initial=initial_,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return signal_box_prefix_codes

    def fetch_prefix_codes(self, initial=None, update=False, dump_dir=None, verbose=False,
                           **kwargs):
        """
        Fetches data of `signal box prefix codes`_.

        .. _`signal box prefix codes`: http://www.railwaycodes.org.uk/signal/signal_boxes0.shtm

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of signal box prefix code;
            defaults to ``None``.
        :type initial: str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of signal box prefix codes and
            the date whey they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_prefix_codes = sb.fetch_prefix_codes()
            >>> type(sb_prefix_codes)
            dict
            >>> list(sb_prefix_codes.keys())
            ['Signal boxes', 'Last updated date']
            >>> sb.KEY
            'Signal boxes'
            >>> sb_prefix_codes_dat = sb_prefix_codes[sb.KEY]
            >>> type(sb_prefix_codes_dat)
            pandas.core.frame.DataFrame
            >>> sb_prefix_codes_dat.head()
              Code               Signal Box  ...            Closed        Control to
            0   AF  Abbey Foregate Junction  ...
            1   AJ           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            2    R           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            3   AW               Abbey Wood  ...      13 July 1975      Dartford (D)
            4   AE         Abbey Works East  ...   1 November 1987  Port Talbot (PT)
            [5 rows x 8 columns]
        """

        if initial:
            args = {
                'data_name': validate_initial(initial),
                'method': self.collect_prefix_codes,
                'sub_dir': "a-z",
                'initial': initial,
            }
            kwargs.update(args)

            signal_box_prefix_codes = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:

            verbose_1 = get_collect_verbosity_for_fetch(data_dir=dump_dir, verbose=verbose)
            verbose_2 = verbose_1 if is_homepage_connectable() else False

            # Get every data table
            data = [
                self.fetch_prefix_codes(initial=x, update=update, verbose=verbose_2)
                for x in string.ascii_lowercase]

            if all(d[x] is None for d, x in zip(data, string.ascii_uppercase)):
                if update:
                    print_instance_connection_error(verbose=verbose)
                    print_void_collection_message(data_name=self.KEY.lower(), verbose=verbose)

                data = [
                    self.fetch_prefix_codes(initial=x, update=False, verbose=verbose_1)
                    for x in string.ascii_lowercase]

            # Select DataFrames only
            signal_boxes_codes_ = (item[x] for item, x in zip(data, string.ascii_uppercase))
            signal_boxes_codes = pd.concat(signal_boxes_codes_, ignore_index=True)

            # Get the latest updated date
            last_updated_dates = (item[self.KEY_TO_LAST_UPDATED_DATE] for item in data)
            latest_update_date = max(d for d in last_updated_dates if d is not None)

            # Create a dict to include all information
            signal_box_prefix_codes = {
                self.KEY: signal_boxes_codes,
                self.KEY_TO_LAST_UPDATED_DATE: latest_update_date
            }

        if dump_dir:
            self._save_data_to_file(
                data=signal_box_prefix_codes, data_name=self.KEY, dump_dir=dump_dir,
                verbose=verbose)

        return signal_box_prefix_codes

    def _collect_non_national_rail_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        non_national_rail_codes = {}

        for h in soup.find_all('h3'):
            # Get the name of the non-national rail
            non_national_rail_name = h.text

            # Find text descriptions
            desc = h.find_next('p')
            desc_text = desc.text.replace('\xa0', '')
            more_desc = desc.find_next('p')
            while more_desc.find_previous('h3') == h:
                desc_text = '\n'.join([desc_text, more_desc.text.replace('\xa0', '')])
                more_desc = more_desc.find_next('p')
                if more_desc is None:
                    break

            # Get table data
            tbl_dat = desc.find_next('table')
            if tbl_dat.find_previous('h3').text == non_national_rail_name:
                ths = [th.text for th in tbl_dat.find_all('th')]  # header
                # trs = tbl_dat.find_next('table').find_all('tr')
                trs = tbl_dat.find_all('tr')
                data = parse_tr(trs=trs, ths=ths, as_dataframe=True)
            else:
                data = None

            # Update data dict
            non_national_rail_codes_ = {
                'Codes': data,
                'Notes': desc_text.replace('\xa0', '').strip(),
            }
            non_national_rail_codes[non_national_rail_name] = non_national_rail_codes_

        non_national_rail_codes_data = {
            self.KEY_TO_NON_NATIONAL_RAIL: non_national_rail_codes,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=non_national_rail_codes_data, data_name=self.KEY_TO_NON_NATIONAL_RAIL,
            verbose=verbose)

        return non_national_rail_codes_data

    def collect_non_national_rail_codes(self, confirmation_required=True, verbose=False,
                                        raise_error=False):
        """
        Collects signal box prefix codes for `non-national rail
        <http://www.railwaycodes.org.uk/signal/signal_boxesX.shtm>`_ from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the signal box prefix codes for non-national rail and
            the date when they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> nnr_codes = sb.collect_non_national_rail_codes()
            To collect data of non-national rail signal box prefix codes
            ? [No]|Yes: yes
            >>> type(nnr_codes)
            dict
            >>> list(nnr_codes.keys())
            ['Non-National Rail', 'Last updated date']
            >>> sb.KEY_TO_NON_NATIONAL_RAIL
            'Non-National Rail'
            >>> nnr_codes_dat = nnr_codes[sb.KEY_TO_NON_NATIONAL_RAIL]
            >>> type(nnr_codes_dat)
            dict
            >>> list(nnr_codes_dat.keys())
            ['Croydon Tramlink signals',
             'Docklands Light Railway signals',
             'Edinburgh Tramway signals',
             'Glasgow Subway signals',
             'London Underground signals',
             'Luas signals',
             'Manchester Metrolink signals',
             'Midland Metro signals',
             'Nottingham Tram signals',
             'Sheffield Supertram signals',
             'Tyne & Wear Metro signals',
             "Heritage, minor and miniature railways and other 'special' signals"]
            >>> lu_signals_codes = nnr_codes_dat['London Underground signals']
            >>> type(lu_signals_codes)
            dict
            >>> list(lu_signals_codes.keys())
            ['Codes', 'Notes']
            >>> type(lu_signals_codes['Codes'])
            pandas.core.frame.DataFrame
            >>> lu_signals_codes['Codes'].head()
              Code  ... Became or taken over by (where known)
            0  BMX  ...                                     -
            1    A  ...                                     -
            2    S  ...                                     -
            3    X  ...                                     -
            4    R  ...                                     -
            [5 rows x 5 columns]
        """

        data_name = self.KEY_TO_NON_NATIONAL_RAIL.lower() + " signal box prefix codes"

        non_national_rail_codes_data = self._collect_data_from_source(
            data_name=data_name, method=self._collect_non_national_rail_codes,
            url=self.catalogue[self.KEY_TO_NON_NATIONAL_RAIL],
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return non_national_rail_codes_data

    def fetch_non_national_rail_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches signal box prefix codes for `non-national rail`_.

        .. _`non-national rail`: http://www.railwaycodes.org.uk/signal/signal_boxesX.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the signal box prefix codes for non-national rail and
            the date when they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> nnr_codes = sb.fetch_non_national_rail_codes()
            >>> type(nnr_codes)
            dict
            >>> list(nnr_codes.keys())
            ['Non-National Rail', 'Last updated date']
            >>> sb.KEY_TO_NON_NATIONAL_RAIL
            'Non-National Rail'
            >>> nnr_codes_dat = nnr_codes[sb.KEY_TO_NON_NATIONAL_RAIL]
            >>> type(nnr_codes_dat)
            dict
            >>> list(nnr_codes_dat.keys())
            ['Croydon Tramlink signals',
             'Docklands Light Railway signals',
             'Edinburgh Tramway signals',
             'Glasgow Subway signals',
             'London Underground signals',
             'Luas signals',
             'Manchester Metrolink signals',
             'Midland Metro signals',
             'Nottingham Tram signals',
             'Sheffield Supertram signals',
             'Tyne & Wear Metro signals',
             "Heritage, minor and miniature railways and other 'special' signals"]
            >>> lu_signals_codes = nnr_codes_dat['London Underground signals']
            >>> type(lu_signals_codes)
            dict
            >>> list(lu_signals_codes.keys())
            ['Codes', 'Notes']
            >>> type(lu_signals_codes['Codes'])
            pandas.core.frame.DataFrame
            >>> lu_signals_codes['Codes'].head()
              Code  ... Became or taken over by (where known)
            0  BMX  ...                                     -
            1    A  ...                                     -
            2    S  ...                                     -
            3    X  ...                                     -
            4    R  ...                                     -
            [5 rows x 5 columns]
        """

        args = {
            'data_name': self.KEY_TO_NON_NATIONAL_RAIL,
            'method': self.collect_non_national_rail_codes,
        }
        kwargs.update(args)

        non_national_rail_codes_data = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return non_national_rail_codes_data

    def _collect_ireland_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        thead, tbody = soup.find('thead'), soup.find('tbody')

        ths = [th.text for th in thead.find_all(name='th')]
        trs = tbody.find_all(name='tr')
        ireland_codes = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        notes_ = soup.find('h4').find_next('ol').find_all('li')
        notes = [li.text for li in notes_]

        ireland_codes_data = {
            self.KEY_TO_IRELAND: ireland_codes,
            'Notes': notes,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=ireland_codes_data, data_name=self.KEY_TO_IRELAND, verbose=verbose)

        return ireland_codes_data

    def collect_ireland_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `Irish signal cabin prefix codes
        <http://www.railwaycodes.org.uk/signal/signal_boxes1.shtm>`_ from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of Irish signal cabin prefix codes and
            the date when they were last updated, or ``None`` if no data is collectd.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> ireland_sb_codes = sb.collect_ireland_codes()
            To collect data of signal box prefix codes of Ireland
            ? [No]|Yes: yes
            >>> type(ireland_sb_codes)
            dict
            >>> list(ireland_sb_codes.keys())
            ['Ireland', 'Notes', 'Last updated date']
            >>> sb.KEY_TO_IRELAND
            'Ireland'
            >>> ireland_sb_codes_dat = ireland_sb_codes[sb.KEY_TO_IRELAND]
            >>> type(ireland_sb_codes_dat)
            pandas.core.frame.DataFrame
            >>> ireland_sb_codes_dat.head()
               Code Signal Cabin                    Note
            0    AD     Adelaide
            1    AN       Antrim
            2    AE      Athlone
            3  AE R                      Distant signals
            4    XG               Level crossing signals
        """

        data_name = "signal box prefix codes of " + self.KEY_TO_IRELAND

        ireland_codes_data = self._collect_data_from_source(
            data_name=data_name, method=self._collect_ireland_codes,
            url=self.catalogue[self.KEY_TO_IRELAND], additional_fields='Notes',
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return ireland_codes_data

    def fetch_ireland_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `Irish signal cabin prefix codes`_.

        .. _`Irish signal cabin prefix codes`:
            http://www.railwaycodes.org.uk/signal/signal_boxes1.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of Irish signal cabin prefix codes and
            the date when they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> ireland_sb_codes = sb.fetch_ireland_codes()
            >>> type(ireland_sb_codes)
            dict
            >>> list(ireland_sb_codes.keys())
            ['Ireland', 'Notes', 'Last updated date']
            >>> sb.KEY_TO_IRELAND
            'Ireland'
            >>> ireland_sb_codes_dat = ireland_sb_codes[sb.KEY_TO_IRELAND]
            >>> type(ireland_sb_codes_dat)
            pandas.core.frame.DataFrame
            >>> ireland_sb_codes_dat.head()
               Code Signal Cabin                    Note
            0    AD     Adelaide
            1    AN       Antrim
            2    AE      Athlone
            3  AE R                      Distant signals
            4    XG               Level crossing signals
        """

        kwargs.update({'data_name': self.KEY_TO_IRELAND, 'method': self.collect_ireland_codes})

        ireland_codes_data = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return ireland_codes_data

    @staticmethod
    def _parse_tbl_dat(h3_or_h4, ths):
        trs = []
        tr = h3_or_h4.find_next('tr')
        while tr:
            if not tr.td.has_attr('colspan'):
                trs.append(tr)
                tr = tr.find_next('tr')
            else:
                break

        tbl = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        return tbl

    def _collect_wr_mas_dates(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        ths = [th.text for th in soup.find('thead').find_all('th')]

        wr_mas_dates = collections.defaultdict(dict)

        for h3 in soup.find_all('h3'):
            h4 = h3.find_next('h4')

            if h4 is not None:
                while h4:
                    prev_h3 = h4.find_previous('h3')
                    if prev_h3.text == h3.text:
                        wr_mas_dates[h3.text].update({h4.text: self._parse_tbl_dat(h4, ths)})
                        h4 = h4.find_next('h4')
                    elif h3.text not in wr_mas_dates.keys():
                        wr_mas_dates.update({h3.text: self._parse_tbl_dat(h3, ths)})
                        break
                    else:
                        break

            else:
                wr_mas_dates.update({h3.text: self._parse_tbl_dat(h3, ths)})

        wr_mas_dates_data = {
            self.KEY_TO_WRMASD: dict(wr_mas_dates),
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=wr_mas_dates_data, data_name=self.KEY_TO_WRMASD, verbose=verbose)

        return wr_mas_dates_data

    def collect_wr_mas_dates(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `WR (western region) MAS (multiple aspect signalling) dates
        <http://www.railwaycodes.org.uk/signal/dates.shtm>`_
        from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of WR MAS dates and
            the date when they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_wr_mas_dates = sb.collect_wr_mas_dates()
            To collect data of WR MAS dates
            ? [No]|Yes: yes
            >>> type(sb_wr_mas_dates)
            dict
            >>> list(sb_wr_mas_dates.keys())
            ['WR MAS dates', 'Last updated date']
            >>> sb.KEY_TO_WRMASD
            'WR MAS dates'
            >>> sb_wr_mas_dates_dat = sb_wr_mas_dates[sb.KEY_TO_WRMASD]
            >>> type(sb_wr_mas_dates_dat)
            dict
            >>> list(sb_wr_mas_dates_dat.keys())[:5]
            ['Paddington-Hayes',
             'Birmingham',
             'Plymouth',
             'Reading-Hayes',
             'Newport Multiple Aspect Signalling']
            >>> sb_wr_mas_dates_dat['Paddington-Hayes']
              Stage             Date                        Area
            0    1A    12 April 1953               Hayes-Hanwell
            1    1B    20 March 1955        Hanwell-Acton Middle
            2    1C  1 February 1959  Acton West-Friars Junction
        """

        wr_mas_dates_data = self._collect_data_from_source(
            data_name=self.KEY_TO_WRMASD, method=self._collect_wr_mas_dates,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return wr_mas_dates_data

    def fetch_wr_mas_dates(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `WR (western region) MAS (multiple aspect signalling) dates`_.

        .. _`WR (western region) MAS (multiple aspect signalling) dates`:
            http://www.railwaycodes.org.uk/signal/dates.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of WR MAS dates and
            the date when they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_wr_mas_dates = sb.fetch_wr_mas_dates()
            >>> type(sb_wr_mas_dates)
            dict
            >>> list(sb_wr_mas_dates.keys())
            ['WR MAS dates', 'Last updated date']
            >>> sb.KEY_TO_WRMASD
            'WR MAS dates'
            >>> sb_wr_mas_dates_dat = sb_wr_mas_dates[sb.KEY_TO_WRMASD]
            >>> type(sb_wr_mas_dates_dat)
            dict
            >>> list(sb_wr_mas_dates_dat.keys())[:5]
            ['Paddington-Hayes',
             'Birmingham',
             'Plymouth',
             'Reading-Hayes',
             'Newport Multiple Aspect Signalling']
            >>> sb_wr_mas_dates_dat['Paddington-Hayes']
              Stage             Date                        Area
            0    1A    12 April 1953               Hayes-Hanwell
            1    1B    20 March 1955        Hanwell-Acton Middle
            2    1C  1 February 1959  Acton West-Friars Junction
        """

        kwargs.update({'data_name': self.KEY_TO_WRMASD, 'method': self.collect_wr_mas_dates})

        wr_mas_dates_data = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return wr_mas_dates_data

    def _collect_bell_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        bell_codes_ = {}
        h3s = soup.find_all('h3')
        for h3 in h3s:
            tbl = h3.find_next('table')
            trs, ths = tbl.find_all('tr'), [th.text for th in tbl.find_all('th')]
            dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

            notes = h3.find_next('p').text

            bell_codes_.update({h3.text: {'Codes': dat, 'Notes': notes}})

        bell_codes = {
            self.KEY_TO_BELL_CODES: bell_codes_,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=bell_codes, data_name=self.KEY_TO_BELL_CODES, verbose=verbose)

        return bell_codes

    def collect_bell_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `bell codes <http://www.railwaycodes.org.uk/signal/bellcodes.shtm>`_
        from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of bell codes and
            the date when they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_bell_codes = sb.collect_bell_codes()
            To collect data of Bell codes
            ? [No]|Yes: yes
            >>> type(sb_bell_codes)
            dict
            >>> list(sb_bell_codes.keys())
            ['Bell codes', 'Last updated date']
            >>> sb.KEY_TO_BELL_CODES
            'Bell codes'
            >>> sb_bell_codes_dat = sb_bell_codes[sb.KEY_TO_BELL_CODES]
            >>> type(sb_bell_codes_dat)
            dict
            >>> list(sb_bell_codes_dat.keys())
            ['Network Rail codes',
             'Southern Railway codes',
             'Lancashire & Yorkshire Railway codes']
            >>> sb_nr_bell_codes = sb_bell_codes_dat['Network Rail codes']
            >>> type(sb_nr_bell_codes)
            dict
            >>> list(sb_nr_bell_codes.keys())
            ['Codes', 'Notes']
            >>> sb_nr_bell_codes_dat = sb_nr_bell_codes['Codes']
            >>> type(sb_nr_bell_codes_dat)
            pandas.core.frame.DataFrame
            >>> sb_nr_bell_codes_dat.head()
                Code                                       Meaning
            0      1                                Call attention
            1    1-1             Answer telephone [withdrawn 2007]
            2  1-1-6           Police assistance urgently required
            3    1-2  Signaller required on telephone [added 2007]
            4  1-2-1                             Train approaching
        """

        bell_codes = self._collect_data_from_source(
            data_name=self.KEY_TO_BELL_CODES, method=self._collect_bell_codes,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return bell_codes

    def fetch_bell_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `bell codes`_.

        .. _`bell codes`: http://www.railwaycodes.org.uk/signal/bellcodes.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of bell codes and
            the date when they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import SignalBoxes  # from pyrcs import SignalBoxes
            >>> sb = SignalBoxes()
            >>> sb_bell_codes = sb.fetch_bell_codes()
            >>> type(sb_bell_codes)
            dict
            >>> list(sb_bell_codes.keys())
            ['Bell codes', 'Last updated date']
            >>> sb.KEY_TO_BELL_CODES
            'Bell codes'
            >>> sb_bell_codes_dat = sb_bell_codes[sb.KEY_TO_BELL_CODES]
            >>> type(sb_bell_codes_dat)
            dict
            >>> list(sb_bell_codes_dat.keys())
            ['Network Rail codes',
             'Southern Railway codes',
             'Lancashire & Yorkshire Railway codes']
            >>> sb_nr_bell_codes = sb_bell_codes_dat['Network Rail codes']
            >>> type(sb_nr_bell_codes)
            dict
            >>> list(sb_nr_bell_codes.keys())
            ['Codes', 'Notes']
            >>> sb_nr_bell_codes_dat = sb_nr_bell_codes['Codes']
            >>> type(sb_nr_bell_codes_dat)
            pandas.core.frame.DataFrame
            >>> sb_nr_bell_codes_dat.head()
                Code                                       Meaning
            0      1                                Call attention
            1    1-1             Answer telephone [withdrawn 2007]
            2  1-1-6           Police assistance urgently required
            3    1-2  Signaller required on telephone [added 2007]
            4  1-2-1                             Train approaching
        """

        kwargs.update({'data_name': self.KEY_TO_BELL_CODES, 'method': self.collect_bell_codes})

        bell_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return bell_codes



"""
Collects codes of `water troughs <http://www.railwaycodes.org.uk/misc/troughs.shtm>`_.
"""

import re
import urllib.parse

import bs4
import numpy as np
import unicodedata

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import homepage_url


def _decode_vulgar_fraction(x):
    for s in x:
        try:
            name = unicodedata.name(s)
            if name.startswith('VULGAR FRACTION'):
                # normalized = unicodedata.normalize('NFKC', s)
                # numerator, _, denominator = normalized.partition('⁄')
                # frac_val = int(numerator) / int(denominator)
                frac_val = unicodedata.numeric(s)
                return frac_val
        except (TypeError, ValueError):
            pass


def _parse_vulgar_fraction_in_length(x):
    """
    Parses 'VULGAR FRACTION' for 'Length' of water trough locations.
    """

    if x == '':
        yd = np.nan

    elif re.match(r'\d+yd', x):  # e.g. '620yd'
        yd = int(re.search(r'\d+(?=yd)', x).group(0))

    elif re.match(r'\d+&frac\d+;yd', x):  # e.g. '506&frac23;yd'
        yd, frac = re.search(r'(\d+)&frac(\d+)(?=;yd)', x).groups()
        yd = int(yd) + int(frac[0]) / int(frac[1])

    else:  # e.g. '557½yd'
        yd = _decode_vulgar_fraction(x)

    return yd


class WaterTroughs(_Base):
    """
    A class for collecting data of
    `water troughs locations <http://www.railwaycodes.org.uk/features/troughs.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Water trough locations'
    #: The key for accessing the data.
    KEY: str = 'Water troughs'
    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/features/troughs.shtm')
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import WaterTroughs  # from pyrcs import WaterTroughs
            >>> wt = WaterTroughs()
            >>> wt.NAME
            'Water trough locations'
        """

        super().__init__(
            data_dir=data_dir, data_category="other-assets", update=update, verbose=verbose)

    def _collect_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        ths = [th.text.strip() for th in soup.find('thead').find_all('th')]
        trs = soup.find('tbody').find_all('tr')

        dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        if 'Length' in dat.columns:
            dat['Length (Yard)'] = dat.Length.map(_parse_vulgar_fraction_in_length)

        water_troughs_codes = {
            self.KEY: dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=water_troughs_codes, data_name=self.KEY, dump_dir=self._cdd("..", "features"),
            verbose=verbose)

        return water_troughs_codes

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects codes of `water troughs locations`_ from the source web page.

        .. _`water troughs locations`: http://www.railwaycodes.org.uk/misc/troughs.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the codes of water trough locations and
            the date they were last updated.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import WaterTroughs  # from pyrcs import WaterTroughs
            >>> wt = WaterTroughs()
            >>> wt_codes = wt.collect_codes()
            To collect data of water troughs
            ? [No]|Yes: yes
            >>> type(wt_codes)
            dict
            >>> list(wt_codes.keys())
            ['Water troughs', 'Last updated date']
            >>> wt.KEY
            'Water troughs'
            >>> wt_codes_dat = wt_codes[wt.KEY]
            >>> type(wt_codes_dat)
            pandas.core.frame.DataFrame
            >>> wt_codes_dat.head()
                ELR  ... Length (Yard)
            0   BEI  ...           NaN
            1   BHL  ...    620.000000
            2  CGJ2  ...      0.666667
            3  CGJ6  ...    561.000000
            4  CGJ6  ...    560.000000
            [5 rows x 6 columns]
        """

        water_troughs_codes = self._collect_data_from_source(
            data_name=self.KEY.lower(), method=self._collect_codes, url=self.URL,
            confirmation_required=confirmation_required,
            confirmation_prompt=f"To collect data of {self.KEY.lower()}\n?",
            verbose=verbose, raise_error=raise_error)

        return water_troughs_codes

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches codes of `water troughs locations`_.

        .. _`water troughs locations`: http://www.railwaycodes.org.uk/misc/troughs.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the codes of water trough locations and
            the date they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import WaterTroughs  # from pyrcs import WaterTroughs
            >>> wt = WaterTroughs()
            >>> wt_codes = wt.fetch_codes()
            >>> type(wt_codes)
            dict
            >>> list(wt_codes.keys())
            ['Water troughs', 'Last updated date']
            >>> wt.KEY
            'Water troughs'
            >>> wt_codes_dat = wt_codes[wt.KEY]
            >>> type(wt_codes_dat)
            pandas.core.frame.DataFrame
            >>> wt_codes_dat.head()
                ELR  ... Length (Yard)
            0   BEI  ...           NaN
            1   BHL  ...    620.000000
            2  CGJ2  ...      0.666667
            3  CGJ6  ...    561.000000
            4  CGJ6  ...    560.000000
            [5 rows x 6 columns]
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_codes,
            'data_dir': self._cdd("..", "features"),
        }
        kwargs.update(args)

        troughs_locations_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return troughs_locations_codes



"""
This module provides the foundational class for managing data persistence, including
saving, loading and fetching data from files. It standardises file naming conventions,
handles directory management and supports verbose logging for debugging.
"""

import copy
import inspect
import os

import pandas as pd
import requests
from pyhelpers._cache import _print_failure_message
from pyhelpers.dirs import cd, validate_dir
from pyhelpers.ops import confirmed, fake_requests_headers
from pyhelpers.store import load_data, save_data

from .parser import get_catalogue, get_introduction, get_last_updated_date
from .utils import cd_data, format_confirmation_prompt, get_collect_verbosity_for_fetch, \
    homepage_url, print_collection_message, print_connection_warning, \
    print_instance_connection_error, print_void_collection_message


class _Base:
    """
    The base class for handling railway codes and related data.
    """

    #: The name of the data.
    NAME: str = 'Railway Codes and other data'
    #: The key for accessing the data.
    KEY: str = ''
    #: The URL of the main web page for the data.
    URL: str = homepage_url()
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, content_type=None, data_category="", data_cluster=None,
                 update=False, verbose=True):
        """
        :param data_dir: Path to the directory where the data is stored; defaults to ``None``.
        :type data_dir: str | None
        :param content_type: Type of content to process, defaults to ``None``.
            Valid options (case-insensitive):

            - ``'catalogue'``, ``'catalog'``, or ``'cat'`` for catalogue-related content.
            - ``'introduction'`` or ``'intro'`` for introductory content.

        :type content_type: str | bool | None
        :param data_category: Category to which the data belongs; defaults to ``''``.
        :type data_category: str
        :param data_cluster: Optional override for ``KEY`` to specify a different subdirectory name.
        :type data_cluster: str | None
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict | None catalogue: Dictionary containing catalogue data
            (if ``content_type='catalogue'``).
        :ivar str | None introduction: Introduction text (if ``content_type='introduction'``).
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The specific directory being used for the current operation.

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> _b = _Base()
            >>> _b.NAME
            'Railway Codes and other data'
            >>> _b.URL
            'http://www.railwaycodes.org.uk/'
        """

        print_connection_warning(verbose=verbose)

        self.catalogue, self.introduction = None, None

        verbose_ = True if verbose == 2 else False

        if isinstance(content_type, str):
            content_type_ = content_type.lower()

            if content_type_.startswith('cat'):  # Get the catalogue of the data
                self.catalogue = get_catalogue(url=self.URL, update=update, verbose=verbose_)

            elif content_type_.startswith('intro'):  # Get the introductory text of the data
                self.introduction = get_introduction(url=self.URL, verbose=verbose_)

        elif content_type:  # Get both the catalogue and introductory text of the data
            self.catalogue = get_catalogue(url=self.URL, update=update, verbose=verbose_)

            self.introduction = get_introduction(url=self.URL, verbose=verbose_)

        self.last_updated_date = get_last_updated_date(url=self.URL)

        # Initialise the data directory for storing or retrieving data
        self.data_dir, self.current_data_dir = self._setup_data_dir(
            data_dir=data_dir, category=data_category, cluster=data_cluster)

    def _setup_data_dir(self, data_dir, category, cluster=None, **kwargs):
        # noinspection PyShadowingNames
        """
        Specifies the initial data directory for a class instance to manage a specific data cluster.

        :param data_dir: The base directory where data files (e.g. pickle files) will be stored.
        :type data_dir: str | None
        :param category: The data category (e.g. ``"line-data"``).
        :type category: str
        :param cluster: Optional override for ``KEY`` to specify a different subdirectory name.
        :type cluster: str | None
        :param kwargs: [Optional] Additional parameters passed to :func:`~pyrcs.utils.cd_data`.
        :return: A tuple containing the default data directory path and
            the current data directory path.
        :rtype: tuple[str, str]

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> import os
            >>> _b = _Base()
            >>> data_dir, current_data_dir = _b._setup_data_dir("data", category="line-data")
            >>> assert data_dir == _b.data_dir
            >>> os.path.relpath(data_dir)
            'data'
            >>> assert current_data_dir == _b.current_data_dir
            >>> os.path.relpath(current_data_dir)
            'data'
        """

        if data_dir:
            self.data_dir = validate_dir(data_dir)

        else:
            cluster_ = self.KEY if cluster is None else copy.copy(cluster)
            self.data_dir = cd_data(category, cluster_.lower().replace(" ", "-"), **kwargs)

        self.current_data_dir = copy.copy(self.data_dir)

        return self.data_dir, self.current_data_dir

    def _cdd(self, *sub_dir, mkdir=True, **kwargs):
        """
        Changes the current directory to the package's data directory,
        or its specified subdirectories (or file).

        The default data directory is: ``"data\\<data_category>\\<data_name>"``.

        :param sub_dir: One or more subdirectories and/or a file to navigate to
            within the data directory.
        :type sub_dir: str
        :param mkdir: Whether to create the specified directory if it doesn't exist;
            defaults to ``True``.
        :type mkdir: bool
        :param kwargs: [Optional] Additional parameters for the `pyhelpers.dir.cd()`_ function.
        :return: The path to the backup data directory or its specified subdirectories (or file).
        :rtype: str

        .. _`pyhelpers.dir.cd()`:
            https://pyhelpers.readthedocs.io/en/latest/_generated/pyhelpers.dir.cd.html

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> import os
            >>> _b = _Base()
            >>> os.path.relpath(_b._cdd())
            'pyrcs\\data'
        """

        path = cd(self.data_dir, *sub_dir, mkdir=mkdir, **kwargs)

        return path

    @staticmethod
    def _format_confirmation_message(data_name, confirmation_prompt=None, initial=None, **kwargs):
        # noinspection PyShadowingNames
        """
        Generates a confirmation prompt message.

        :param confirmation_prompt: A custom confirmation message
            or a callable function that generates one.
        :type confirmation_prompt: str | typing.Callable | None
        :param data_name: The name of the data being confirmed
            (passed to the prompt function if callable).
        :type data_name: str
        :return: The generated confirmation prompt as a string.
        :rtype: str

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> from pyrcs.utils import format_confirmation_prompt
            >>> _b = _Base()
            >>> data_name = '"test_data_name"'
            >>> prompt = _b._format_confirmation_message(data_name)
            >>> prompt
            'To collect data of "test_data_name"\n?'
            >>> prompt = _b._format_confirmation_message(data_name, format_confirmation_prompt)
            >>> prompt
            'To collect data of "test_data_name"\n?'
            >>> prompt = _b._format_confirmation_message(data_name, "test message")
            >>> prompt
            'test message'
            >>> prompt = _b._format_confirmation_message(data_name, format_confirmation_prompt, 'a')
            >>> prompt
            'To collect data of "test_data_name" beginning with "a"\n?'
        """

        if confirmation_prompt:
            if isinstance(confirmation_prompt, str):
                prompt = confirmation_prompt
            else:
                assert callable(confirmation_prompt)
                if 'initial' in inspect.signature(confirmation_prompt).parameters:
                    kwargs.update({'initial': initial})
                prompt = confirmation_prompt(data_name, **kwargs)

        else:
            prompt = format_confirmation_prompt(data_name=data_name, initial=initial)

        return prompt

    def _fallback_data(self, key=None, additional_fields=None):
        """
        Returns a fallback data dictionary in case of failure.
        """

        if not key:
            return None

        data = {key: None, self.KEY_TO_LAST_UPDATED_DATE: None}

        if additional_fields:
            if isinstance(additional_fields, dict):
                data.update(additional_fields)
            else:
                data.update({additional_fields: None})

        return data

    def _collect_data_from_source(self, data_name, method, url=None, initial=None,
                                  additional_fields=None, confirmation_required=True,
                                  confirmation_prompt=None, verbose=False, raise_error=False,
                                  **kwargs):
        """
        Collects and parses data from a specified source webpage.

        :param data_name: The descriptive name of the data being collected
            (used for lookups and messages).
        :type data_name: str | None
        :param method: The parsing function to execute upon successful data retrieval.
            The function **must** accept:

            - ``source`` (*requests.Response*): The HTTP response object.
            - ``verbose`` (*bool | int*): The verbosity flag.

            The function **may** optionally accept:

            - ``data_name`` (*str*): Injected automatically if present in the function signature.
            - ``initial`` (*str*): Injected automatically if present in the function signature.
            - Any other arguments passed via ``**kwargs``.

        :type method: typing.Callable
        :param url: The target URL. If ``None``, the method attempts to retrieve the URL from
            ``self.catalogue`` using ``initial`` or ``data_name`` as the key.
        :type url: str | None
        :param initial: The initial letter/code used to categorize the data
            (e.g. 'A' for stations starting with A); it is used as a fallback key for URL lookup
            if ``url`` is not provided.
        :type initial: str | None
        :param additional_fields: Key-value pairs to include in the fallback dictionary
            if data collection fails; it is useful for ensuring consistent data structure
            (e.g. returning ``None`` for specific columns).
        :type additional_fields: dict | str | None
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param confirmation_prompt: A custom message or a callable that generates a message
            for the confirmation prompt.
        :type confirmation_prompt: str | typing.Callable | None
        :param verbose: Whether to print status messages and errors to the console;
            defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: If ``True``, exceptions (network or parsing) are raised to the caller;
            if ``False`` (default), exceptions are caught and fallback data is returned.
        :type raise_error: bool
        :param kwargs: [Optional] Additional keyword arguments passed directly to the ``method``.
        :return: The data returned by ``method``, or a dictionary containing fallback values
            (e.g. ``{key: None}``) on failure.
        :rtype: pandas.DataFrame | dict | None
        :raises ValueError: If the URL cannot be resolved and ``raise_error=True``.
        :raises requests.RequestException: If a network error occurs and ``raise_error=True``.

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> _b = _Base()
            >>> _b.catalogue = {'A': 'https://github.com/mikeqfu/pyrcs'}
            >>> _b._collect_data_from_source("test_data_name", method=_b._fallback_data)
        """

        # Confirmation step
        prompt = self._format_confirmation_message(
            data_name=data_name, confirmation_prompt=confirmation_prompt, initial=initial)

        if not confirmed(prompt=prompt, confirmation_required=confirmation_required):
            return None

        print_collection_message(
            data_name=data_name, initial=initial, verbose=verbose,
            confirmation_required=confirmation_required)

        # Prepare fallback data
        fallback_data = self._fallback_data(key=initial, additional_fields=additional_fields)

        # Resolve URL
        target_url = url or self.catalogue.get(initial or data_name)

        if not target_url:
            if initial:
                err_msg = f'No data is available for codes beginning with "{initial}".'
            else:  # defaults to data_name context
                err_msg = \
                    f'"{data_name}" not found in `.catalogue`. Check `.catalogue` for valid keys.'

            if raise_error:
                raise ValueError(err_msg)
            elif verbose:
                print(err_msg)
            return fallback_data

        # Fetch and process
        try:
            # Network request
            source = requests.get(url=target_url, headers=fake_requests_headers(), timeout=30)
            source.raise_for_status()  # Raises HTTPError for bad responses

            # Dynamic argument injection
            collector_kwargs = kwargs.copy()
            collector_kwargs.update({'source': source, 'verbose': verbose})

            # Inspect the collector method to see if it requires extra arguments
            params = inspect.signature(method).parameters
            if 'data_name' in params:
                collector_kwargs['data_name'] = data_name
            if 'initial' in params:
                collector_kwargs['initial'] = initial

            # Execute Parsing Method
            data = method(**collector_kwargs)

            return data

        except requests.RequestException as e:  # Handle network/HTTP errors
            print_instance_connection_error(verbose=verbose, e=e, raise_error=raise_error)
            return fallback_data

        except Exception as e:  # Handle parsing/method errors
            _print_failure_message(e, "Failed. Error:", verbose=verbose, raise_error=raise_error)
            return fallback_data

    def _make_file_pathname(self, data_name, ext=".pkl", data_dir=None, sub_dir=None, **kwargs):
        """
        Generates a standardised file pathname for saving data.

        The method constructs a filename by converting the given ``data_name`` to lowercase
        and replacing spaces with hyphens. The filename is then appended with the specified
        file extension (``ext``). If a ``data_dir`` is provided, the file will be placed in that
        directory; otherwise, the default directory within the class will be used.

        :param data_name: The key identifying the data within a specific code cluster.
        :type data_name: str
        :param ext: The file extension (including the leading dot), defaults to ``".pkl"``.
        :type ext: str
        :param data_dir: The directory where the file will be saved; defaults to ``None``.
        :type data_dir: str | os.PathLike | None
        :param sub_dir: A subdirectory name or a list of subdirectory names; defaults to ``None``.
        :type sub_dir: str | list | None
        :return: The pathname for saving the data file.
        :rtype: str

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> import os
            >>> _b = _Base()
            >>> pathname = _b._make_file_pathname("example-data", ext=".pkl")
            >>> os.path.relpath(pathname)
            'pyrcs\\data\\example-data.pkl'
            >>> pathname = _b._make_file_pathname("example-data", ext=".json", sub_dir="a-z")
            >>> os.path.relpath(pathname)
            'pyrcs\\data\\a-z\\example-data.json'
        """

        filename = data_name.lower().replace(",", "").replace(" ", "-") + ext

        if sub_dir:
            sub_dir_ = [sub_dir] if isinstance(sub_dir, str) else sub_dir
        else:
            sub_dir_ = []

        if data_dir:
            self.current_data_dir = validate_dir(path_to_dir=None if data_dir is True else data_dir)
            file_pathname = os.path.join(self.current_data_dir, *sub_dir_, filename)

        else:  # data_dir is None or data_dir == ""
            file_pathname = self._cdd(*sub_dir_, filename, **kwargs)

        return file_pathname

    def _save_data_to_file(self, data, data_name, ext=".pkl", dump_dir=None, sub_dir=None,
                           verbose=False, **kwargs):
        # noinspection PyShadowingNames
        """
        Saves the provided ``data`` to a file using the specified format and location.

        This method determines the appropriate file format and saves the ``data`` accordingly.
        If ``dump_dir`` is specified, the file will be saved there; otherwise, a default directory
        managed by the class is used. If ``verbose`` is set to ``True`` or ``2``, status messages
        will be printed to the console.

        :param data: The data to be saved.
        :type data: pandas.DataFrame | list | dict
        :param data_name: A unique identifier for the data.
        :type data_name: str
        :param ext: The file extension or a boolean indicating whether to save the data;
            defaults to ``".pkl"``.
        :type ext: str | bool
        :param dump_dir: The directory where the file should be saved;
            if ``None`` (default) a default directory within the class is used.
        :type dump_dir: str | os.PathLike | None
        :param sub_dir: A subdirectory name or a list of subdirectory names; defaults to ``None``.
        :type sub_dir: str | list | None
        :param verbose: Whether to print detailed information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param kwargs: [Optional] Additional parameters passed to `pyhelpers.store.save_data()`_.

        .. _`pyhelpers.store.save_data()`:
            https://pyhelpers.readthedocs.io/en/latest/_generated/pyhelpers.store.save_data.html

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> from pyhelpers._cache import example_dataframe
            >>> import os
            >>> _b = _Base()
            >>> data = example_dataframe()
            >>> _b._save_data_to_file(data, data_name="Example data", ext=".csv", verbose=2)
            Saving "example-data.csv" to ".\\pyrcs\\data\\" ... Done.
            >>> example_filepath = "pyrcs\\data\\example-data.csv"
            >>> os.path.isfile(example_filepath)
            True
            >>> os.remove(example_filepath)
        """

        # Check if data has content
        data_has_contents = bool(data) if not isinstance(data, pd.DataFrame) else True

        if data_has_contents:
            # Ensure extension starts with '.'
            if isinstance(ext, str):
                file_ext = "." + ext if not ext.startswith(".") else copy.copy(ext)
            else:
                file_ext = ".pkl"

            path_to_file = self._make_file_pathname(
                data_name=data_name, ext=file_ext, data_dir=dump_dir, sub_dir=sub_dir)

            save_data(data=data, path_to_file=path_to_file, verbose=(verbose == 2), **kwargs)

        else:
            print_void_collection_message(data_name=data_name, verbose=verbose)

    def _fetch_data_from_file(self, data_name, method, ext=".pkl", update=False, dump_dir=None,
                              verbose=False, raise_error=False, data_dir=None, sub_dir=None,
                              save_data_kwargs=None, **kwargs):
        # noinspection PyShadowingNames
        """
        Fetches data from a stored file or generates it using the specified ``method``.

        This method attempts to load data from a backup file based on the provided parameters.
        If the file exists and ``update=False``, the data is loaded from the file. Otherwise, the
        ``method`` is called to generate the data, which can then be optionally saved to a specified
        directory (``dump_dir``).

        :param data_name: A unique identifier for the data, used to determine the filename.
        :type data_name: str
        :param method: A callable method that generates the data if the file does not exist
            or an update is required.
        :type method: typing.Callable
        :param ext: The file extension or a boolean indicating whether to save the data;
            defaults to ``".pkl"``.
        :type ext: bool | str
        :param update: Whether to perform an update check on the package data;
            defaults to ``False``.
        :type update: bool
        :param dump_dir: The directory where the file is stored or should be saved;
            defaults to ``None``.
        :type dump_dir: str | pathlib.Path | None
        :param verbose: Whether to print detailed information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param data_dir: The directory where the data should be fetched from;
            if ``None``, the default data directory is used.
        :type data_dir: str | os.PathLike | None
        :param sub_dir: A subdirectory name or a list of subdirectory names; defaults to ``None``.
        :type sub_dir: str | list | None
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :param save_data_kwargs: [Optional] Additional parameters for the
            :func:`pyrcs.utils.save_data_to_file` function; defaults to ``None``.
        :type save_data_kwargs: dict | None
        :param kwargs: [Optional] Additional parameters passed to ``method`` when generating data.
        :return: The fetched or generated data;
            returns ``None`` if an error occurs and  ``raise_error=False``.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs._base import _Base
            >>> from pyhelpers._cache import example_dataframe
            >>> import os
            >>> _b = _Base()
            >>> def create_example_dataframe(confirmation_required=False, verbose=False):
            ...     assert isinstance(confirmation_required, bool)
            ...     assert isinstance(verbose, (bool, int))
            ...     return example_dataframe()
            >>> data = _b._fetch_data_from_file("example-data", create_example_dataframe)
            >>> data
                        Longitude   Latitude
            City
            London      -0.127647  51.507322
            Birmingham  -1.902691  52.479699
            Manchester  -2.245115  53.479489
            Leeds       -1.543794  53.797418
        """

        try:
            # Generate the file path
            path_to_file = self._make_file_pathname(
                data_name=data_name, ext=ext, data_dir=data_dir, sub_dir=sub_dir)

            if os.path.isfile(path_to_file) and not update:  # Attempt to load existing data
                data = load_data(path_to_file, verbose=(verbose == 2))

            else:
                verbose_ = get_collect_verbosity_for_fetch(data_dir=dump_dir, verbose=verbose)

                kwargs.update({'confirmation_required': False, 'verbose': verbose_})

                if isinstance(method, str):
                    data = getattr(self, method)(**kwargs)
                else:
                    data = method(**kwargs)

            if dump_dir:
                self._save_data_to_file(
                    data=data, data_name=data_name, ext=ext, dump_dir=dump_dir, sub_dir=sub_dir,
                    verbose=verbose, **(save_data_kwargs or {}))

            return data

        except Exception as e:
            _print_failure_message(e=e, prefix="Error:", verbose=verbose, raise_error=raise_error)



"""
Collects `telegraph codes <http://www.railwaycodes.org.uk/misc/telegraph.shtm>`_.
"""

import urllib.parse

import bs4

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import homepage_url


def _parse_telegraph_in_use_term(x):
    if x == '♠':
        y = 'cross industry term used in 1939'

    elif x == '†':
        y = 'cross industry term used in 1939 and still used by BR in the 1980s'

    else:
        y = x

    return y


class Telegraph(_Base):
    """
    A class for collecting data of
    `telegraph code words <http://www.railwaycodes.org.uk/features/telegraph.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Telegraph code words'
    #: The key for accessing the data.
    KEY: str = 'Telegraphic codes'
    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/features/telegraph.shtm')
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Telegraph  # from pyrcs import Telegraph
            >>> tel = Telegraph()
            >>> tel.NAME
            'Telegraph code words'
        """

        super().__init__(
            data_dir=data_dir, data_category="other-assets", update=update, verbose=verbose)

    def _collect_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        h3s = soup.find_all('h3')

        sub_keys, codes_list = [], []
        for h3 in h3s:
            sub_keys.append(h3.text.strip())

            ths = [th.text.strip() for th in h3.find_next('thead').find_all('th')]
            trs = h3.find_next('tbody').find_all('tr')

            dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

            if 'In use' in dat.columns:
                dat['In use'] = dat['In use'].map(_parse_telegraph_in_use_term)

            codes_list.append(dat)

        telegraph_code_words_dat = dict(zip(sub_keys, codes_list))

        telegraph_code_words = {
            self.KEY: telegraph_code_words_dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=telegraph_code_words, data_name=self.KEY, dump_dir=self._cdd("..", "features"),
            verbose=verbose)

        return telegraph_code_words

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `telegraph code words`_ from the source web page.

        .. _`telegraph code words`: http://www.railwaycodes.org.uk/misc/telegraph.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of telegraph code words and
            the date they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import Telegraph  # from pyrcs import Telegraph
            >>> tel = Telegraph()
            >>> tel_codes = tel.collect_codes()
            To collect data of telegraphic codes
            ? [No]|Yes: yes
            >>> type(tel_codes)
            dict
            >>> list(tel_codes.keys())
            ['Telegraphic codes', 'Last updated date']
            >>> tel.KEY
            'Telegraphic codes'
            >>> tel_codes_dat = tel_codes[tel.KEY]
            >>> type(tel_codes_dat)
            dict
            >>> list(tel_codes_dat.keys())
            ['Official codes', 'Unofficial codes']
            >>> tel_official_codes = tel_codes_dat['Official codes']
            >>> type(tel_official_codes)
            pandas.core.frame.DataFrame
            >>> tel_official_codes.head()
                  Code  ...                               In use
            0    ABACK  ...     cross industry term used in 1939
            1    ABASE  ...                            GWR, 1939
            2  ABREAST  ...  GWR, 1939 / Railway Executive, 1950
            3  ABREAST  ...   British Transport Commission, 1958
            4   ABSENT  ...                            GWR, 1939
            [5 rows x 3 columns]
            >>> tel_unofficial_codes = tel_codes_dat['Unofficial codes']
            >>> type(tel_unofficial_codes)
            pandas.core.frame.DataFrame
            >>> tel_unofficial_codes.head()
                  Code                             Unofficial description
            0  CRANKEX                                      [See KRANKEX]
            1  DRUNKEX  Saturday night special train (usually a DMU) t...
            2     GYFO    Strongly urge all speed ('Get your finger out')
            3  KRANKEX  Special train with interesting routing or trac...
            4   MYSTEX  Special excursion going somewhere no one reall...
        """

        telegraph_code_words = self._collect_data_from_source(
            data_name=self.KEY.lower(), method=self._collect_codes, url=self.URL,
            confirmation_required=confirmation_required,
            confirmation_prompt=f"To collect data of {self.KEY.lower()}\n?",
            verbose=verbose, raise_error=raise_error)

        return telegraph_code_words

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `telegraph code words`_.

        .. _`telegraph code words`: http://www.railwaycodes.org.uk/misc/telegraph.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of telegraph code words and
            the date they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Telegraph  # from pyrcs import Telegraph
            >>> tel = Telegraph()
            >>> tel_codes = tel.fetch_codes()
            >>> type(tel_codes)
            dict
            >>> list(tel_codes.keys())
            ['Telegraphic codes', 'Last updated date']
            >>> tel.KEY
            'Telegraphic codes'
            >>> tel_codes_dat = tel_codes[tel.KEY]
            >>> type(tel_codes_dat)
            dict
            >>> list(tel_codes_dat.keys())
            ['Official codes', 'Unofficial codes']
            >>> tel_official_codes = tel_codes_dat['Official codes']
            >>> type(tel_official_codes)
            pandas.core.frame.DataFrame
            >>> tel_official_codes.head()
                  Code  ...                               In use
            0    ABACK  ...     cross industry term used in 1939
            1    ABASE  ...                            GWR, 1939
            2  ABREAST  ...  GWR, 1939 / Railway Executive, 1950
            3  ABREAST  ...   British Transport Commission, 1958
            4   ABSENT  ...                            GWR, 1939
            [5 rows x 3 columns]
            >>> tel_unofficial_codes = tel_codes_dat['Unofficial codes']
            >>> type(tel_unofficial_codes)
            pandas.core.frame.DataFrame
            >>> tel_unofficial_codes.head()
                  Code                             Unofficial description
            0  CRANKEX                                      [See KRANKEX]
            1  DRUNKEX  Saturday night special train (usually a DMU) t...
            2     GYFO    Strongly urge all speed ('Get your finger out')
            3  KRANKEX  Special train with interesting routing or trac...
            4   MYSTEX  Special excursion going somewhere no one reall...
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_codes,
            'data_dir': self._cdd("..", "features"),
        }
        kwargs.update(args)

        telegraph_code_words = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return telegraph_code_words



"""
Provides a number of helper functions.
"""

import importlib.resources
import math
import os
import re
import string

import pandas as pd
from pyhelpers._cache import _format_error_message, _print_failure_message
from pyhelpers.ops import confirmed, is_url_connectable
from pyhelpers.store import load_data, save_data


# == Specify address of web pages ==================================================================


def homepage_url():
    """
    Returns the homepage URL of the data source.

    :return: The homepage URL of the data source.
    :rtype: str

    **Examples**::

        >>> from pyrcs.utils import homepage_url
        >>> homepage_url()
        'http://www.railwaycodes.org.uk/'
    """

    return 'http://www.railwaycodes.org.uk/'


# == Validate inputs ===============================================================================


def is_homepage_connectable():
    """
    Checks and returns whether the Railway Codes website is reacheable.

    :return: Whether the Railway Codes website is reacheable.
    :rtype: bool

    **Examples**::

        >>> from pyrcs.utils import is_homepage_connectable
        >>> is_homepage_connectable()
        True
    """

    url = homepage_url()

    rslt = is_url_connectable(url=url)

    return rslt


def is_str_float(x, finite_only=False):
    """
    Checks if a string represents and can be converted to a finite float value.

    This function attempts to convert the input to a float. It returns ``False``
    for non-numeric strings, ``None`` inputs, or special floating point
    values like ``NaN`` or ``Infinity`` (optional).

    :param x: String-type data or any object that can be converted to float.
    :type x: str | typing.Any
    :param finite_only: Whether to return ``False`` for special values such as
        ``NaN`` and ``Infinity``; defaults to ``False``.
    :type finite_only: bool
    :return: ``True`` if the string represents a valid finite number, ``False`` otherwise.
    :rtype: bool

    **Examples**::

        >>> from pyrcs.utils import is_str_float
        >>> is_str_float('')
        False
        >>> is_str_float('a')
        False
        >>> is_str_float('1')
        True
        >>> is_str_float('1.1')
        True
        >>> is_str_float('nan', finite_only=True)
        False
        >>> is_str_float('inf')
        True
    """

    # Handle NoneType or non-string inputs explicitly if needed
    if x is None:
        return False

    try:
        y = float(x)  # float(re.sub('[()~]', '', text))

        if finite_only:  # Reject NaN and Infinity a real number is needed
            return math.isfinite(y)

        return True

    except (ValueError, TypeError):
        return False


def validate_initial(initial, as_is=False):
    """
    Validates if a value is a single ASCII letter.

    :param initial: The input value to validate,
        which is expected to be a string representing a single letter.
    :type initial: str
    :param as_is: If ``as_is=True``, the function returns the letter in its original case;
        if ``as_is=False`` (default), the letter is returned in uppercase.
    :type as_is: bool
    :return: The validated initial letter.
    :rtype: str
    :raises ValueError: If the input is not a single ASCII letter.
    :raises TypeError: If the input is not a string.

    **Examples**::

        >>> from pyrcs.utils import validate_initial
        >>> validate_initial('x')
        'X'
        >>> validate_initial('x', as_is=True)
        'x'
        >>> validate_initial('xyz')
        Traceback (most recent call last):
          ...
            ...
        ValueError: 'xyz' is invalid; it must be a single letter (A-Z, a-z).
    """

    # Type check
    if not isinstance(initial, str):
        raise TypeError(f"`initial` must be a string, not {type(initial).__name__}.")

    # Validation check (using string.ascii_letters)
    if initial not in set(string.ascii_letters):
        raise ValueError(f"'{initial}' is invalid; it must be a single letter (A-Z, a-z).")

    return initial if as_is else initial.upper()


def validate_page_name(cls_instance, page_no, valid_page_no):
    """
    Retrieves the valid page name corresponding to a given page number.

    This method checks if the provided ``page_no`` is within the set of
    valid page numbers. If valid, it returns the name of the page
    associated with the specified ``page_no`` from the catalogue of the given ``cls_instance``.

    :param cls_instance: An instance of the class that contains the page catalogue.
    :type cls_instance: object
    :param page_no: The page number to validate,
        which can be an integer or a string representation of a number.
    :type page_no: int | str
    :param valid_page_no: A collection of valid page numbers,
        which can be a set, list, or tuple containing the allowable page numbers.
    :type valid_page_no: set | list | tuple | range
    :return: The validated page name associated with the given ``page_no`` in the class's catalogue.
    :rtype: str

    .. seealso::

        Examples for the methods:

        - :py:meth:`Tunnels.collect_codes()<pyrcs.other_assets.tunnel.Tunnels.collect_codes>`
        - :py:meth:`Viaducts.collect_codes()<pyrcs.other_assets.viaduct.Viaducts.collect_codes>`
    """

    assert page_no in valid_page_no, f"Valid `page_no` must be one of {valid_page_no}."

    assert hasattr(cls_instance, 'catalogue'), \
        "The input `cls_instance` must have an attribute named `catalogue`."
    # noinspection PyUnresolvedReferences
    page_name = [k for k in cls_instance.catalogue.keys() if str(page_no) in k][0]

    return page_name


def get_collect_verbosity_for_fetch(data_dir, verbose):
    """
    Creates a new parameter that indicates whether to print relevant information to the console.

    This function is used only if it is necessary to re-collect data when trying to fetch the data.

    :param data_dir: The directory path where data is saved.
    :type data_dir: str | None
    :param verbose: The requested verbosity level.
    :type verbose: bool | int
    :return: The resolved verbosity level (boolean or integer).
    :rtype: bool | int

    **Examples**::

        >>> from pyrcs.utils import get_collect_verbosity_for_fetch
        >>> get_collect_verbosity_for_fetch(data_dir="data", verbose=True)
        False
    """

    # If saving to disk (data_dir) or verbose is off, return False.
    if data_dir or not verbose:
        return False

    # Otherwise, preserve the specific integer level (e.g. 2) or default to True.
    return 2 if verbose == 2 else True


def get_batch_fetch_verbosity(data_dir, verbose):
    """
    Creates a new parameter that indicates whether to print relevant information to the console.

    This function is used only when fetching all data of a cluster.

    :param data_dir: Name of the folder where the pickle file is to be saved.
    :type data_dir: str | None
    :param verbose: Whether to print relevant information to the console.
    :type verbose: bool | int
    :return: A boolean indicating whether to print relevant information to the console when
        fetching all data of a cluster.
    :rtype: bool | int

    **Examples**::

        >>> from pyrcs.utils import get_batch_fetch_verbosity
        >>> get_batch_fetch_verbosity(data_dir="data", verbose=True)
        False
    """

    if is_homepage_connectable():
        verbose_ = get_collect_verbosity_for_fetch(data_dir=data_dir, verbose=verbose)
    else:
        verbose_ = False

    return verbose_


# == Print messages ================================================================================


def format_confirmation_prompt(data_name, initial=None, ending="\n?"):
    # noinspection PyShadowingNames
    """
    Returns a message for comfirming whether to proceed to collect a certain cluster of data.

    :param data_name: The name of the dataset to be collected, e.g. ``"Railway Codes"``.
    :type data_name: str
    :param initial: The initial letter for the code; defaults to ``None``.
    :type initial: str | None
    :param ending: The ending of the confirmation message; defaults to ``"\\n?"``.
    :type ending: str
    :return: A confirmation message asking whether to proceed with the dataset collection.
    :rtype: str

    **Examples**::

        >>> from pyrcs.utils import format_confirmation_prompt
        >>> prompt = format_confirmation_prompt(data_name="Railway Codes")
        >>> print(prompt)
        To collect data of Railway Codes
        ?
        >>> prompt = format_confirmation_prompt(data_name="location codes", initial="A")
        >>> print(prompt)
        To collect data of location codes beginning with "A"
        ?
    """

    prompt = f"To collect data of {data_name}"
    if initial:
        if initial.lower() in string.ascii_letters:
            prompt += f' beginning with "{initial}"{ending}'
        else:
            prompt += f' ({initial})'
    else:
        prompt += ending

    return prompt


def print_collection_message(data_name, initial=None, verbose=False, confirmation_required=True,
                             end=" ... "):
    """
    Prints a message indicating the status of data collection.

    :param data_name: The name of the data being collected.
    :type data_name: str
    :param initial: The initial letter of the desired code or data; defaults to ``None``.
    :type initial: str | None
    :param verbose: Whether to print relevant information to the console.
    :type verbose: bool | int
    :param confirmation_required: Whether user confirmation is required before proceeding.
    :type confirmation_required: bool
    :param end: String appended at the end of the message; defaults to ``" ... "``.
    :type end: str

    **Examples**::

        >>> from pyrcs.utils import print_collection_message
        >>> print_collection_message("Railway Codes", verbose=True, confirmation_required=False)
        Collecting the data of Railway Codes ...
    """

    message_ = "Collecting the data"

    if verbose in {True, 1}:
        if confirmation_required:
            print(message_, end=end)

        else:
            message_ += f" of {data_name}"
            if initial:
                if initial.lower() in set(string.ascii_letters):
                    print(f'{message_} beginning with "{initial}"', end=end)
                else:
                    print(f'{message_} ({initial})', end=end)
            else:
                print(message_, end=end)


def print_connection_warning(verbose=False):
    """
    Checks the Internet connection and prints a warning if the source is unreachable.

    This function attempts to connect to the source homepage of Railway Codes.
    If the connection fails and ``verbose`` is ``True``, it alerts the user that the instance
    will rely on local backup data.

    :param verbose: Whether to print the warning message; defaults to ``False``.
    :type verbose: bool | int

    **Examples**::

        >>> from pyrcs.utils import print_connection_warning
        >>> # If the website is reachable, nothing is printed:
        >>> print_connection_warning(verbose=True)

        >>> # If the website is unreachable, a warning is printed:
        >>> print_connection_warning(verbose=True)
        Failed to establish an Internet connection. The current instance relies on local backup.
    """

    if not is_homepage_connectable():
        if verbose:
            print("Failed to establish an Internet connection. "
                  "The current instance relies on local backup.")


def print_instance_connection_error(update=False, verbose=False, e=None, raise_error=False):
    """
    Prints an error message when an instance fails to establish an Internet connection.

    :param update: Indicates whether the error occurred during a data update; defaults to ``False``.
    :type update: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param e: An optional exception message to display.
    :type e: Exception | None
    :param raise_error: Whether to raise the exception;
        if ``raise_error=False`` (default), the error will be suppressed.
    :type raise_error: bool

    **Examples**::

        >>> from pyrcs.utils import print_instance_connection_error
        >>> print_instance_connection_error(verbose=True)
        The Internet connection is not available.
        >>> print_instance_connection_error(update=True, verbose=True)
        The Internet connection is not available. Failed to update the data.
        >>> print_instance_connection_error(update=True, verbose=2, raise_error=True)
        Failed. The Internet connection is not available. Failed to update the data.
        Traceback (most recent call last):
          ...
            ...
        TypeError: exceptions must derive from BaseException
    """

    if not verbose:
        return None  # No need to print anything if verbosity is off

    if verbose == 2:
        print("Failed.", end=" ")

    err_msg = _format_error_message(e) if e else "The Internet connection is not available."

    if update:
        err_msg += " Failed to update the data."

    _print_failure_message(
        e if raise_error else err_msg, prefix="", verbose=True, raise_error=raise_error)


def print_void_collection_message(data_name, verbose):
    """
    Prints a warning message when the data collection process fails and
    no fresh data was collected.

    :param data_name: Name of the data being collected.
    :type data_name: str
    :param verbose: Whether to print relevant information to the console.
    :type verbose: bool | int

    **Examples**::

        >>> from pyrcs.utils import print_void_collection_message
        >>> print_void_collection_message(data_name="Railway Codes", verbose=True)
        No data of "Railway Codes" has been freshly collected.
    """

    if verbose:
        print(f"No data of \"{data_name.title()}\" has been freshly collected.")


# == Save and retrieve pre-packed data =============================================================


def cd_data(*sub_dir, data_dir="data", mkdir=False, **kwargs):
    """
    Specifies or changes to a directory (or subdirectories) for storing backup data for the package.

    :param sub_dir: Name(s) of directories and/or a filename to navigate to or create.
    :type sub_dir: str
    :param data_dir: The base directory for storing data; defaults to ``"data"``.
    :type data_dir: str
    :param mkdir: Whether to create the specified directory if it does not exist;
        defaults to ``False``.
    :type mkdir: bool
    :param kwargs: [Optional] Additional parameters for `os.makedirs()`_, such as ``mode=0o777``.
    :return: The full path of the specified directory or file within ``data_dir``.
    :rtype: str

    .. _`os.makedirs()`: https://docs.python.org/3/library/os.html#os.makedirs

    **Examples**::

        >>> from pyrcs.utils import cd_data
        >>> import os
        >>> path_to_dat_dir = cd_data(data_dir="data")
        >>> os.path.relpath(path_to_dat_dir)
        'pyrcs\\data'
    """

    path = importlib.resources.files(__package__).joinpath(data_dir)
    for x in sub_dir:
        path = os.path.join(path, x)

    if mkdir:
        path_to_file, ext = os.path.splitext(path)
        if ext == '':
            os.makedirs(path_to_file, exist_ok=True, **kwargs)
        else:
            os.makedirs(os.path.dirname(path_to_file), exist_ok=True, **kwargs)

    return path


def fetch_location_names_errata(k=None, regex=False, as_dataframe=False, column_name=None):
    """
    Fetches a dictionary or dataframe to rectify location names.

    :param k: The key for the errata dictionary; defaults to ``None``.
    :type k: str | int | float | bool | None
    :param regex: Whether to create the dictionary for replacements based on regular expressions;
        defaults to ``False``.
    :type regex: bool
    :param as_dataframe: Whether to return the dictionary as a dataframe; defaults to ``False``.
    :type as_dataframe: bool
    :param column_name: If ``as_dataframe=True``, the column name for the dataframe;
        defaults to ``None``.
    :type column_name: str | list | None
    :return: A dictionary for rectifying location names, or a dataframe if requested.
    :rtype: dict | pandas.DataFrame

    **Examples**::

        >>> from pyrcs.utils import fetch_location_names_errata
        >>> repl_dict = fetch_location_names_errata()
        >>> type(repl_dict)
        dict
        >>> list(repl_dict.keys())[:5]
        ['"Tyndrum Upper" (Upper Tyndrum)',
         'AISH EMERGENCY CROSSOVER',
         'ATLBRJN',
         'Aberdeen Craiginches',
         'Aberdeen Craiginches T.C.']
        >>> repl_dict = fetch_location_names_errata(regex=True, as_dataframe=True)
        >>> type(repl_dict)
        pandas.core.frame.DataFrame
        >>> repl_dict.head()
                                         new_value
        re.compile(' \\(DC lines\\)')   [DC lines]
        re.compile(' And | \\+ ')               &
        re.compile('-By-')                    -by-
        re.compile('-In-')                    -in-
        re.compile('-En-Le-')              -en-le-
    """

    json_filename = "location-names-errata{}.json".format("" if not regex else "-regex")
    location_name_repl_dict = load_data(cd_data(json_filename))

    if regex:
        location_name_repl_dict = {re.compile(k): v for k, v in location_name_repl_dict.items()}

    replacement_dict = {k: location_name_repl_dict} if k else location_name_repl_dict

    if as_dataframe:
        if column_name is None:
            col_name = ['Name']
        else:
            col_name = [column_name] if isinstance(column_name, str) else column_name.copy()
        replacement_dict = pd.DataFrame.from_dict(replacement_dict, orient='index', columns=col_name)

    return replacement_dict


def _update_location_names_errata(new_items, regex, verbose=False):
    """
    Updates the location name replacement dictionary in the package data.

    :param new_items: The new items to add or replace in the dictionary.
    :type new_items: dict
    :param regex: Whether this update applies to the regular expression dictionary.
    :type regex: bool
    :param verbose: Whether to print detailed information to the console; defaults to ``False``.
    :type verbose: bool | int
    """

    json_filename = "location-names-errata{}.json".format("" if not regex else "-regex")

    new_items_keys = list(new_items.keys())

    if confirmed(f'To update "{json_filename}" with {{"{new_items_keys[0]}"... }}?'):
        path_to_json = cd_data(json_filename)
        location_name_repl_dict = load_data(path_to_json)

        if any(isinstance(k, re.Pattern) for k in new_items_keys):
            new_items = {k.pattern: v for k, v in new_items.items() if isinstance(k, re.Pattern)}

        location_name_repl_dict.update(new_items)

        save_data(location_name_repl_dict, path_to_json, indent=4, verbose=verbose)

    else:
        if verbose:
            print("Cancelled.")



"""
Collects British `railway track diagrams <http://www.railwaycodes.org.uk/track/diagrams0.shtm>`_.
"""

import urllib.parse

import bs4
import pandas as pd
from pyhelpers._cache import _print_failure_message

from .._base import _Base
from ..parser import _get_last_updated_date
from ..utils import cd_data, homepage_url


class TrackDiagrams(_Base):
    """
    A class for collecting data of British
    `railway track diagrams <http://www.railwaycodes.org.uk/track/diagrams0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway track diagrams'
    #: The key for accessing the data.
    KEY: str = 'Track diagrams'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/line/diagrams0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the data catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.line_data import TrackDiagrams  # from pyrcs import TrackDiagrams
            >>> td = TrackDiagrams()
            >>> td.NAME
            'Railway track diagrams'
            >>> td.URL
            'http://www.railwaycodes.org.uk/line/diagrams0.shtm'
        """

        super().__init__(
            data_dir=data_dir, data_category="line-data", update=update, verbose=verbose)

        self.catalogue = self.fetch_catalogue(update=update, verbose=(verbose == 2 or False))

    def _collect_catalogue(self, source, verbose=False):
        track_diagrams_catalogue_ = {}

        try:
            soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

            h3 = soup.find('h3', string=True, attrs={'class': None})
            while h3:
                # Description
                if h3.text == 'Miscellaneous':
                    desc = [x.text for x in h3.find_next_siblings('p')]
                else:
                    desc = h3.find_next_sibling('p').text.replace('\xa0', '')

                # Extract details
                cold_soup = h3.find_next('div', attrs={'class': 'columns'})
                if cold_soup:
                    info = [x.text for x in cold_soup.find_all('p') if x.string != '\xa0']
                    urls = [
                        urllib.parse.urljoin(self.URL, a.get('href'))
                        for a in cold_soup.find_all('a')]
                else:
                    cold_soup = h3.find_next('a', attrs={'target': '_blank'})
                    info, urls = [], []

                    while cold_soup:
                        info.append(cold_soup.text)
                        urls.append(urllib.parse.urljoin(self.URL, cold_soup['href']))
                        if h3.text == 'Miscellaneous':
                            cold_soup = cold_soup.find_next('a')
                        else:
                            cold_soup = cold_soup.find_next_sibling('a')

                meta = pd.DataFrame(data=zip(info, urls), columns=['Description', 'FileURL'])

                track_diagrams_catalogue_.update({h3.text: (desc, meta)})

                h3 = h3.find_next_sibling('h3')

            track_diagrams_catalogue = {
                self.KEY: track_diagrams_catalogue_,
                self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
            }

            if verbose in {True, 1}:
                print("Done.")

            self._save_data_to_file(
                data_name=self.KEY.lower(), data=track_diagrams_catalogue,
                dump_dir=cd_data("catalogue"), verbose=verbose)

            return track_diagrams_catalogue

        except Exception as e:
            _print_failure_message(e)

    def collect_catalogue(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects the catalogue of sample `railway track diagrams`_ from the source web page.

        .. _`railway track diagrams`: http://www.railwaycodes.org.uk/track/diagrams0.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the railway track diagram catalogue and
            the date it was last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import TrackDiagrams  # from pyrcs import TrackDiagrams
            >>> td = TrackDiagrams()
            >>> track_diagrams_catalog = td.collect_catalogue()
            To collect the catalogue of track diagrams
            ? [No]|Yes: yes
            >>> type(track_diagrams_catalog)
            dict
            >>> list(track_diagrams_catalog.keys())
            ['Track diagrams', 'Last updated date']
            >>> td_dat = track_diagrams_catalog['Track diagrams']
            >>> type(td_dat)
            dict
            >>> list(td_dat.keys())
            ['Main line diagrams', 'Tram systems', 'London Underground', 'Miscellaneous']
            >>> main_line_diagrams = td_dat['Main line diagrams']
            >>> type(main_line_diagrams)
            tuple
            >>> type(main_line_diagrams[1])
            pandas.core.frame.DataFrame
            >>> main_line_diagrams[1].head()
                                         Description                                         FileURL
            0  South Central area (1985) 10.4Mb file  http://www.railwaycodes.org.uk/line/track/d...
            1   South Eastern area (1976) 5.4Mb file  http://www.railwaycodes.org.uk/line/track/d...
        """

        track_diagrams_catalogue = self._collect_data_from_source(
            data_name=self.KEY.lower(), method=self._collect_catalogue, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return track_diagrams_catalogue

    def fetch_catalogue(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches the catalogue of `railway track diagrams`_.

        .. _`railway track diagrams`: http://www.railwaycodes.org.uk/track/diagrams0.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the catalogue of railway track diagrams and
            the date it was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import TrackDiagrams  # from pyrcs import TrackDiagrams
            >>> td = TrackDiagrams()
            >>> trk_diagr_cat = td.fetch_catalogue()
            >>> type(trk_diagr_cat)
            dict
            >>> list(trk_diagr_cat.keys())
            ['Track diagrams', 'Last updated date']
            >>> td_dat = trk_diagr_cat['Track diagrams']
            >>> type(td_dat)
            dict
            >>> list(td_dat.keys())
            ['Main line diagrams', 'Tram systems', 'London Underground', 'Miscellaneous']
            >>> main_line_diagrams = td_dat['Main line diagrams']
            >>> type(main_line_diagrams)
            tuple
            >>> type(main_line_diagrams[1])
            pandas.core.frame.DataFrame
            >>> main_line_diagrams[1].head()
                                         Description                                         FileURL
            0  South Central area (1985) 10.4Mb file  http://www.railwaycodes.org.uk/line/track/d...
            1   South Eastern area (1976) 5.4Mb file  http://www.railwaycodes.org.uk/line/track/d...
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_catalogue,
            'data_dir': cd_data("catalogue"),
        }
        kwargs.update(args)

        track_diagrams_catalogue = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return track_diagrams_catalogue



"""
Collects data of railway codes.

.. note::

    The current release only includes
    `line data <http://www.railwaycodes.org.uk/linedatamenu.shtm>`_ and
    `other assets <http://www.railwaycodes.org.uk/otherassetsmenu.shtm>`_.
"""

import time

from pyhelpers.ops import confirmed

from .line_data import Bridges, ELRMileages, Electrification, LOR, LineNames, LocationIdentifiers, \
    TrackDiagrams
from .other_assets import Buzzer, Depots, Features, HabdWild, SignalBoxes, Stations, Telegraph, \
    Tunnels, Viaducts, WaterTroughs
from .parser import get_category_menu
from .utils import is_homepage_connectable, print_connection_warning, print_instance_connection_error


class _Base:
    #: The name of the data.
    NAME: str = 'Railway Codes and other data'

    def __init__(self, update=False, verbose=True, raise_error=False):
        if not is_homepage_connectable():
            self.connected = False
            print_connection_warning(verbose=verbose)

        else:
            self.connected = True

        self.cls_init_kwargs = {'update': update, 'verbose': verbose}

        self.catalogue = get_category_menu(
            self.NAME, update=update, confirmation_required=False, verbose=verbose,
            raise_error=raise_error)


class LineData(_Base):
    """
    A class representation of all modules of the subpackage :py:mod:`~pyrcs.line_data`
    for collecting the codes of `line data`_.

    .. _`line data`: http://www.railwaycodes.org.uk/linedatamenu.shtm
    """

    #: The name of the data.
    NAME: str = 'Line data'

    def __init__(self, update=False, verbose=True, raise_error=False):
        """
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool

        :ivar bool connected: Whether the Internet / the Railway Codes website is connected.
        :ivar dict catalogue: The catalogue of the line data.
        :ivar ELRMileages ELRMileages: An instance of the :class:`~elr_mileage.ELRMileages` class.
        :ivar Electrification Electrification:
            An instance of the :class:`~elec.Electrification` class.
        :ivar LocationIdentifiers LocationIdentifiers:
            An instance of the :class:`~loc_id.LocationIdentifiers` class.
        :ivar LOR LOR: An instance of the :class:`~lor_code.LOR` class.
        :ivar LineNames LineNames: An instance of the :class:`~line_name.LineNames` class.
        :ivar TrackDiagrams TrackDiagrams:
            An instance of the :class:`~trk_diagr.TrackDiagrams` class.
        :ivar Bridges Bridges: An instance of the :py:class:`~bridge.Bridges` class.

        **Examples**::

            >>> from pyrcs import LineData
            >>> ld = LineData()
            >>> # To get data of location codes
            >>> location_codes = ld.LocationIdentifiers.fetch_codes()
            >>> type(location_codes)
            dict
            >>> list(location_codes.keys())
            ['LocationID', 'Other systems', 'Additional notes', 'Last updated date']
            >>> location_codes_dat = location_codes[ld.LocationIdentifiers.KEY]
            >>> type(location_codes_dat)
            pandas.core.frame.DataFrame
            >>> location_codes_dat.head()
                                           Location CRS  ... STANME_Note STANOX_Note
            0                                Aachen      ...
            1                    Abbeyhill Junction      ...
            2                 Abbeyhill Signal E811      ...
            3            Abbeyhill Turnback Sidings      ...
            4  Abbey Level Crossing (Staffordshire)      ...
            [5 rows x 12 columns]
            >>> # To get data of line names
            >>> line_names_codes = ld.LineNames.fetch_codes()
            >>> type(line_names_codes)
            dict
            >>> list(line_names_codes.keys())
            ['Line names', 'Last updated date']
            >>> line_names_codes_dat = line_names_codes[ld.LineNames.KEY]
            >>> type(line_names_codes_dat)
            pandas.core.frame.DataFrame
            >>> line_names_codes_dat.head()
                         Line name  ... Route_note
            0           Abbey Line  ...       None
            1        Airedale Line  ...       None
            2          Argyle Line  ...       None
            3     Arun Valley Line  ...       None
            4  Atlantic Coast Line  ...       None
            [5 rows x 3 columns]
        """

        super().__init__(update=update, verbose=verbose, raise_error=raise_error)

        # Relevant classes
        self.ELRMileages = ELRMileages(**self.cls_init_kwargs)
        self.Electrification = Electrification(**self.cls_init_kwargs)
        self.LocationIdentifiers = LocationIdentifiers(**self.cls_init_kwargs)
        self.LOR = LOR(**self.cls_init_kwargs)
        self.LineNames = LineNames(**self.cls_init_kwargs)
        self.TrackDiagrams = TrackDiagrams(**self.cls_init_kwargs)
        self.Bridges = Bridges(**self.cls_init_kwargs)

    def update(self, confirmation_required=True, verbose=False, interval=5, init_update=False):
        """
        Updates the pre-packed `line data`_.

        .. _`line data`: http://www.railwaycodes.org.uk/linedatamenu.shtm

        :param confirmation_required: Whether user confirmation is required before proceeding;
            defaults to ``True``.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param interval: A time gap (in seconds) between the updating of different classes,
            defaults to ``5``
        :type interval: int or float
        :param init_update: Whether to update the data for each subclass when being instantiated,
            defaults to ``False``
        :type init_update: bool

        **Examples**::

            >>> from pyrcs.collector import LineData
            >>> ld = LineData()
            >>> ld.update(verbose=True)
        """

        if not self.connected:
            print_instance_connection_error(verbose=verbose)

        else:
            if confirmed("To update line data\n?", confirmation_required=confirmation_required):

                if init_update:
                    self.__init__(update=init_update)

                update_args = {'update': True, 'verbose': verbose}

                # ELR and mileages
                print(f"\n{self.ELRMileages.NAME}:")
                _ = self.ELRMileages.fetch_elr(**update_args)

                time.sleep(interval)

                # Electrification
                print(f"\n{self.Electrification.NAME}:")
                _ = self.Electrification.get_independent_lines_catalogue(**update_args)
                _ = self.Electrification.fetch_codes(**update_args)

                time.sleep(interval)

                # Location
                print(f"\n{self.LocationIdentifiers.NAME}:")
                _ = self.LocationIdentifiers.fetch_codes(**update_args)

                time.sleep(interval)

                # Line of routes
                print(f"\n{self.LOR.NAME}:")
                _ = self.LOR.get_keys_to_prefixes(prefixes_only=True, **update_args)
                _ = self.LOR.get_keys_to_prefixes(prefixes_only=False, **update_args)
                _ = self.LOR.get_page_urls(**update_args)
                _ = self.LOR.fetch_codes(**update_args)
                _ = self.LOR.fetch_elr_lor_converter(**update_args)

                time.sleep(interval)

                # Line names
                print(f"\n{self.LineNames.NAME}:")
                _ = self.LineNames.fetch_codes(**update_args)

                time.sleep(interval)

                # Track diagrams
                print(f"\n{self.TrackDiagrams.NAME}:")
                _ = self.TrackDiagrams.fetch_catalogue(**update_args)

                time.sleep(interval)

                # Bridges
                print(f"\n{self.Bridges.NAME}:")
                _ = self.Bridges.fetch_codes(**update_args)


class OtherAssets(_Base):
    """
    A class representation of all modules of the subpackage :py:mod:`~pyrcs.other_assets`
    for collecting the codes of `other assets`_.

    .. _`other assets`: http://www.railwaycodes.org.uk/otherassetsmenu.shtm
    """

    #: The name of the data.
    NAME: str = 'Other assets'

    def __init__(self, update=False, verbose=True, raise_error=False):
        """
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool

        :ivar bool connected: Whether the Internet / the Railway Codes website is connected.
        :ivar dict catalogue: The catalogue of the data.
        :ivar SignalBoxes SignalBoxes: An instance of the :class:`~sig_box.SignalBoxes` class.
        :ivar Tunnels Tunnels: An instance of the :class:`~tunnel.Tunnels` class.
        :ivar Viaducts Viaducts: An instance of the :class:`~viaduct.Viaducts` class.
        :ivar Stations Stations: An instance of the :class:`~station.Stations` class.
        :ivar Depots Depots: An instance of the :class:`~depot.Depots` class.
        :ivar Features Features: An instance of the :class:`~feature.Features` class.
        :ivar HabdWild HabdWild: An instance of the :class:`~habd_wild.HabdWild` class.
        :ivar WaterTroughs WaterTroughs: An instance of the :class:`~trough.WaterTroughs` class.
        :ivar Telegraph Telegraph: An instance of the :class:`~telegraph.Telegraph` class.
        :ivar Buzzer Buzzer: An instance of the :class:`~buzzer.Buzzer` class.

        **Examples**::

            >>> from pyrcs import OtherAssets
            >>> oa = OtherAssets()
            >>> # To get data of railway stations
            >>> rail_stn_locations = oa.Stations.fetch_locations()
            >>> type(rail_stn_locations)
            dict
            >>> list(rail_stn_locations.keys())
            ['Mileages, operators and grid coordinates', 'Last updated date']
            >>> rail_stn_locations_dat = rail_stn_locations[oa.Stations.KEY_TO_STN]
            >>> type(rail_stn_locations_dat)
            pandas.core.frame.DataFrame
            >>> rail_stn_locations_dat.head()
                       Station  ...                                    Former Operator
            0       Abbey Wood  ...  London & South Eastern Railway from 1 April 20...
            1       Abbey Wood  ...
            2             Aber  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            3        Abercynon  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            4  Abercynon North  ...  [Cardiff Railway Company from 13 October 1996 ...
            [5 rows x 13 columns]
            >>> # To get data of signal boxes
            >>> signal_boxes_codes = oa.SignalBoxes.fetch_prefix_codes()
            >>> type(signal_boxes_codes)
            dict
            >>> list(signal_boxes_codes.keys())
            ['Signal boxes', 'Last updated date']
            >>> signal_boxes_codes_dat = signal_boxes_codes[oa.SignalBoxes.KEY]
            >>> type(signal_boxes_codes_dat)
            pandas.core.frame.DataFrame
            >>> signal_boxes_codes_dat.head()
              Code               Signal Box  ...            Closed        Control to
            0   AF  Abbey Foregate Junction  ...
            1   AJ           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            2    R           Abbey Junction  ...  16 February 1992     Nuneaton (NN)
            3   AW               Abbey Wood  ...      13 July 1975      Dartford (D)
            4   AE         Abbey Works East  ...   1 November 1987  Port Talbot (PT)
            [5 rows x 8 columns]
        """

        super().__init__(update=update, verbose=verbose, raise_error=raise_error)

        # Relevant classes
        self.SignalBoxes = SignalBoxes(**self.cls_init_kwargs)
        self.Tunnels = Tunnels(**self.cls_init_kwargs)
        self.Viaducts = Viaducts(**self.cls_init_kwargs)
        self.Stations = Stations(**self.cls_init_kwargs)
        self.Depots = Depots(**self.cls_init_kwargs)
        self.HabdWild = HabdWild(**self.cls_init_kwargs)
        self.WaterTroughs = WaterTroughs(**self.cls_init_kwargs)
        self.Telegraph = Telegraph(**self.cls_init_kwargs)
        self.Buzzer = Buzzer(**self.cls_init_kwargs)
        self.Features = Features(**self.cls_init_kwargs)

    def update(self, confirmation_required=True, verbose=False, interval=5, init_update=False):
        """
        Updates the pre-packed data of the `other assets`_.

        .. _`other assets`: http://www.railwaycodes.org.uk/otherassetsmenu.shtm

        :param confirmation_required: Whether user confirmation is required before proceeding;
            defaults to ``True``.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param interval: A time gap (in seconds) between the updating of different classes,
            defaults to ``5``
        :type interval: int or float
        :param init_update: Whether to update the data for each subclass when being instantiated,
            defaults to ``False``
        :type init_update: bool

        **Examples**::

            >>> from pyrcs.collector import OtherAssets
            >>> oa = OtherAssets()
            >>> oa.update(verbose=True)
        """

        if not self.connected:
            print_instance_connection_error(verbose=verbose)

        else:
            if confirmed("To update data of other assets\n?", confirmation_required):

                if init_update:
                    self.__init__(update=init_update)

                update_args = {'update': True, 'verbose': verbose}

                # Signal boxes
                print(f"\n{self.SignalBoxes.NAME}:")
                _ = self.SignalBoxes.fetch_prefix_codes(**update_args)
                _ = self.SignalBoxes.fetch_non_national_rail_codes(**update_args)
                _ = self.SignalBoxes.fetch_ireland_codes(**update_args)
                _ = self.SignalBoxes.fetch_wr_mas_dates(**update_args)
                _ = self.SignalBoxes.fetch_bell_codes(**update_args)

                time.sleep(interval)

                # Tunnels
                print(f"\n{self.Tunnels.NAME}:")
                _ = self.Tunnels.fetch_codes(**update_args)

                time.sleep(interval)

                # Viaducts
                print(f"\n{self.Viaducts.NAME}:")
                _ = self.Viaducts.fetch_codes(**update_args)

                time.sleep(interval)

                # Stations
                print(f"\n{self.Stations.NAME}:")
                _ = self.Stations.fetch_catalogue(**update_args)
                _ = self.Stations.fetch_locations(**update_args)

                time.sleep(interval)

                # Depots
                print(f"\n{self.Depots.NAME}:")
                _ = self.Depots.fetch_codes(**update_args)

                time.sleep(interval)

                # Features
                print(f"\n{self.Features.NAME}:")
                _ = self.Features.fetch_codes(**update_args)



"""
Collects `driver/guard buzzer codes <http://www.railwaycodes.org.uk/misc/buzzer.shtm>`_.
"""

import urllib.parse

from .._base import _Base
from ..parser import _get_last_updated_date, parse_table
from ..utils import homepage_url


class Buzzer(_Base):
    """
    A class for collecting data of
    `buzzer codes <http://www.railwaycodes.org.uk/features/buzzer.shtm>`_.
    """

    #: The name of the data.
    NAME = 'Buzzer codes'
    #: The key for accessing the data.
    KEY = 'Buzzer codes'
    #: The URL of the main web page for the data.
    URL = urllib.parse.urljoin(homepage_url(), '/features/buzzer.shtm')
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Buzzer  # from pyrcs import Buzzer
            >>> buz = Buzzer()
            >>> buz.NAME
            'Buzzer codes'
        """

        super().__init__(
            data_dir=data_dir, data_category="other-assets", update=update, verbose=verbose)

    def _collect_codes(self, source, verbose=False):
        codes_dat, soup = parse_table(source=source, parser='html.parser', as_dataframe=True)

        column_names = []
        for col in codes_dat.columns:
            col_name = col.split('\r\n')
            if len(col_name) > 1:
                column_names.append(col_name[0] + ' [' + ''.join(col_name[1:]) + ']')
            else:
                column_names.append(col_name[0])

        codes_dat.columns = column_names

        buzzer_codes = {
            self.KEY: codes_dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=buzzer_codes, data_name=self.KEY, dump_dir=self._cdd("..", "features"),
            verbose=verbose)

        return buzzer_codes

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `buzzer codes`_ from the source web page.

        .. _`buzzer codes`: http://www.railwaycodes.org.uk/misc/buzzer.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of buzzer codes and
            the date they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import Buzzer  # from pyrcs import Buzzer
            >>> buz = Buzzer()
            >>> buz_codes = buz.collect_codes()
            To collect data of Buzzer codes
            ? [No]|Yes: yes
            >>> type(buz_codes)
            dict
            >>> list(buz_codes.keys())
            ['Buzzer codes', 'Last updated date']
            >>> buz.KEY
            'Buzzer codes'
            >>> buz_codes_dat = buz_codes[buz.KEY]
            >>> type(buz_codes_dat)
            pandas.core.frame.DataFrame
            >>> buz_codes_dat.head()
              Code [number of buzzes or groups separated by pauses]            Meaning
            0                                                  1                  Stop
            1                                                1-2           Close doors
            2                                                  2        Ready to start
            3                                                2-2     Do not open doors
            4                                                  3              Set back
        """

        buzzer_codes = self._collect_data_from_source(
            data_name=self.KEY.lower(), method=self._collect_codes, url=self.URL,
            confirmation_required=confirmation_required,
            confirmation_prompt=f"To collect data of {self.KEY.lower()}\n?",
            verbose=verbose, raise_error=raise_error)

        return buzzer_codes

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `buzzer codes`_.

        .. _`buzzer codes`: http://www.railwaycodes.org.uk/misc/buzzer.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of buzzer codes and
            the date they were last updated
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Buzzer  # from pyrcs import Buzzer
            >>> buz = Buzzer()
            >>> buz_codes = buz.fetch_codes()
            >>> type(buz_codes)
            dict
            >>> list(buz_codes.keys())
            ['Buzzer codes', 'Last updated date']
            >>> buz.KEY
            'Buzzer codes'
            >>> buz_codes_dat = buz_codes[buz.KEY]
            >>> type(buz_codes_dat)
            pandas.core.frame.DataFrame
            >>> buz_codes_dat.head()
              Code [number of buzzes or groups separated by pauses]            Meaning
            0                                                  1                  Stop
            1                                                1-2           Close doors
            2                                                  2        Ready to start
            3                                                2-2     Do not open doors
            4                                                  3              Set back
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_codes,
            'data_dir': self._cdd("..", "features"),
        }
        kwargs.update(args)

        buzzer_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return buzzer_codes
